/**
 * 
 */
package de.ibh.StrongRoom.DB;

import de.ibh.StrongRoom.Crypto.CryptoBackend.ECryptoPGP_KeyType;
import de.ibh.StrongRoom.Exception.SRDBSQLException;
import de.ibh.StrongRoom.Exception.SRRoleKeyMismatchException;
import de.ibh.StrongRoom.Exception.SRRoleNotFoundException;
import de.ibh.StrongRoom.Log.DebugMessage;
import de.ibh.StrongRoom.Log.ErrorMessage;
import de.ibh.StrongRoom.Log.ErrorMessage.EMessageTypes;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;



/**
 * @author thomas
 *
 */
public class DBConn {
	private Connection   conn;
	private ErrorMessage errmsg = null;
	private String       cURL = null;
	static private DBConn db = null;
	private final StrongRoomL10N  l10n = StrongRoomL10N.getInstance();
		
	static DebugMessage logger = DebugMessage.getInstance();
	
	static public DBConn getInstance(String cName, String cURL, ErrorMessage errmsg) throws InstantiationException, IllegalAccessException, ClassNotFoundException, SQLException {
		if (db != null)
			return db;

		return db = new DBConn(cName, cURL, errmsg);
	}
	
	static public DBConn getInstance() {
		return db;
	}
	
	public DBConn(String cName, String cURL, ErrorMessage errmsg) throws InstantiationException, IllegalAccessException, ClassNotFoundException, SQLException {
	    this.errmsg = errmsg;
	    this.cURL = cURL;
	    
		// load JDBC-ODBC driver
		Class.forName(cName).newInstance();
		conn = DriverManager.getConnection(this.cURL);
	}
	
	
	private void errorMessage(String msg, EMessageTypes type, Exception e) {
	    if(errmsg != null) errmsg.printErrorMessage(msg, type);
	    if(e != null) logger.debug(e);
	}
	
	
	private boolean reconnect() {
	    //disconnect();

	    try {
            conn = DriverManager.getConnection(this.cURL);
        } catch (SQLException e) {
            errorMessage(l10n._("Failed to reconnect to SQL Server."), EMessageTypes.TYPE_ERROR, e);

            return false;
        }
        
        return true;
	}
	
	
	private boolean handleSQLException(SQLException e) {
	    // Check SQLState
	    if(e.getSQLState().equals("08S01"))
	        return reconnect();
	    else if(e.getSQLState().equals("HY010"))
	        return reconnect();
	    
	    return false;
	}
	
	
	public boolean isConnected() {
        return conn == null ? false : true;
	}

	public synchronized void disconnect() {
	    try {
	    	conn.close();
	    } catch (SQLException e) {
	        errorMessage(l10n._("Can't disconnect from SQL Server."), EMessageTypes.TYPE_ERROR, e);
	    }
        finally {
    	    db = null;
		}
	}

	public synchronized List<Customer> getCustomers() {
		List<Customer> custs = new LinkedList<Customer>();
		
		try {
			PreparedStatement st = conn.prepareStatement("SELECT ID, Name FROM Customers ORDER BY Name ASC");
			try {
				if(st.execute()) {
					ResultSet rs = st.getResultSet();
					try {
						while (rs.next()) {
							custs.add(new Customer(new Long(rs.getString("ID")), rs.getString("Name")));
						}
					}
					finally {
						rs.close();
					}
				}
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error at retrieving list of customers."), EMessageTypes.TYPE_ERROR, e);
			
			return null;
		}
		
		return custs;
	}
	
	
    public synchronized CryptoEntry getCryptoEntry(final Customer cust, final long ID) throws SRRoleKeyMismatchException, SRRoleNotFoundException {
        CryptoEntry ce = null;
        Map<Long, Role> roles = getRoles();
        if(roles == null)
            return null;

        try {
            PreparedStatement st = conn.prepareStatement("SELECT RoleID, RKeyFPR, ID, Subject, Description, Exportable, Readonly, Hidden, UKeyFPR, SignRKey, SignUKey FROM CryptoEntries WHERE ID = ?");
            try {
                st.setLong(1, ID);

                if(st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                        while(rs.next()) {
                            long roleID = new Long(rs.getString("RoleID"));
               
                            if(!roles.containsKey(roleID))
                                throw new SRRoleNotFoundException(roleID);

                            Role role = roles.get(roleID);
                            
                            final String rKeyFPR = rs.getString("RKeyFPR");
                            if(!role.getRoleKey().getKeyFPR().equals(rKeyFPR))
                                throw new SRRoleKeyMismatchException(roleID, rKeyFPR, role.getRoleKey().getKeyFPR());
                        
                            ce = new CryptoEntry(
                                new Long(rs.getString("ID")), // TODO: Change it back to .getLong("...") if the bug is found.
                                cust,
                                rs.getString("Subject"),
                                rs.getString("Description"),
                                role,
                                rKeyFPR,
                                rs.getBoolean("Exportable"),
                                rs.getBoolean("Readonly"),
                                rs.getBoolean("Hidden"),
                                rs.getString("UKeyFPR"),
                                rs.getString("SignRKey"),
                                rs.getString("SignUKey"));
                        }
                    }
                    finally {
                        rs.close();
                    }
                }
            }
            finally {
                st.close();
            }
        } catch (SQLException e) {
            handleSQLException(e);
            errorMessage(l10n._("Error while retrieving crypto entries."), EMessageTypes.TYPE_ERROR, e);
        }
        
        return ce;
    }
	
	
	
	public synchronized List<CryptoEntry> getCryptoEntries(final Customer cust) throws SRRoleKeyMismatchException, SRRoleNotFoundException {
		LinkedList<CryptoEntry> ces = new LinkedList<CryptoEntry>();
		Map<Long, Role> roles = getRoles();
        if(roles == null)
            return null;

        try {
    	    PreparedStatement st = conn.prepareStatement("SELECT RoleID, RKeyFPR, ID, Subject, Description, Exportable, Readonly, Hidden, UKeyFPR, SignRKey, SignUKey FROM CryptoEntries WHERE CustomerID = ? ORDER BY Subject");
        	try {
	            st.setLong(1, cust.getId());

	            if(st.execute()) {
	            	ResultSet rs = st.getResultSet();
	            	try {
	            		while(rs.next()) {
	            			long roleID = new Long(rs.getString("RoleID"));
	           
	            			if(!roles.containsKey(roleID))
	            				throw new SRRoleNotFoundException(roleID);

	            			Role role = roles.get(roleID);
	            			
	            			final String rKeyFPR = rs.getString("RKeyFPR");
	            			if(!role.getRoleKey().getKeyFPR().equals(rKeyFPR))
	            				throw new SRRoleKeyMismatchException(roleID, rKeyFPR, role.getRoleKey().getKeyFPR());
                        
	            			ces.add(new CryptoEntry(
                                new Long(rs.getString("ID")), // TODO: Change it back to .getLong("...") if the bug is found.
                                cust,
                                rs.getString("Subject"),
                                rs.getString("Description"),
                                role,
                                rKeyFPR,
                                rs.getBoolean("Exportable"),
                                rs.getBoolean("Readonly"),
                                rs.getBoolean("Hidden"),
                                rs.getString("UKeyFPR"),
                                rs.getString("SignRKey"),
                                rs.getString("SignUKey"))
	            			);
	            		}
	            	}
	            	finally {
	            		rs.close();
	            	}
	            }
        	}
        	finally {
        		st.close();
        	}
    	} catch (SQLException e) {
    	    handleSQLException(e);
    		errorMessage(l10n._("Error while retrieving crypto entries."), EMessageTypes.TYPE_ERROR, e);

    		return null;
    	}
    	
		return ces;
	}

	public synchronized List<CryptoEntry> getCryptoEntriesByRole(final Role role) throws SRRoleKeyMismatchException, SRDBSQLException {
		final List<Customer> custs = getCustomers();
		final Map<Long, Customer> custMap = new HashMap<Long, Customer>();

		for(Customer cust : custs) {
			custMap.put(cust.getId(), cust);
		}
		
		LinkedList<CryptoEntry> ces = new LinkedList<CryptoEntry>();
        try {
    	    PreparedStatement st = conn.prepareStatement("SELECT RKeyFPR, ID, CustomerID, Subject, Description, Exportable, Readonly, Hidden, UKeyFPR, SignRKey, SignUKey FROM CryptoEntries WHERE RoleID = ?");
        	try {
	            st.setLong(1, role.getID());

	            if(st.execute()) {
	            	ResultSet rs = st.getResultSet();
	            	try {
	            		while(rs.next()) {
	            			final String rKeyFPR = rs.getString("RKeyFPR");
	            			if(!role.getRoleKey().getKeyFPR().equals(rKeyFPR))
	            				throw new SRRoleKeyMismatchException(role.getID(), rKeyFPR, role.getRoleKey().getKeyFPR());
                        
	            			ces.add(new CryptoEntry(
                                new Long(rs.getString("ID")), // TODO: Change it back to .getLong("...") if the bug is found.
                                custMap.get(rs.getLong("CustomerID")),
                                rs.getString("Subject"),
                                rs.getString("Description"),
                                role,
                                rKeyFPR,
                                rs.getBoolean("Exportable"),
                                rs.getBoolean("Readonly"),
                                rs.getBoolean("Hidden"),
                                rs.getString("UKeyFPR"),
                                rs.getString("SignRKey"),
                                rs.getString("SignUKey"))
	            			);
	            		}
	            	}
	            	finally {
	            		rs.close();
	            	}
	            }
        	}
        	finally {
        		st.close();
        	}
    	} catch (SQLException e) {
    	    handleSQLException(e);
    		throw new SRDBSQLException(e);
    	}
    	
		return ces;
	}	
	
	public synchronized List<CryptoData> getCryptoData(final CryptoEntry ce) {
		LinkedList<CryptoData> cds = new LinkedList<CryptoData>();

		try {
			PreparedStatement st = conn.prepareStatement("SELECT TStamp, EncData, UKeyFPR, SignRKey, SignUKey FROM CryptoData WHERE CryptoEntryID = ? ORDER BY TStamp DESC");
			try {
				st.setLong(1, ce.getId());
			
				if (st.execute()) {
					ResultSet rs = st.getResultSet(); 
					try {
						while (rs.next()) {
							cds.add(new CryptoData(
								ce,
								rs.getTimestamp("TStamp"),
								rs.getString("EncData"),
								rs.getString("UKeyFPR"),
								rs.getString("SignRKey"),
								rs.getString("SignUKey"))
							);
						}
					}
					finally {
						rs.close();
					}
				}
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
			errorMessage(l10n._("Error while retrieving crypto data."), EMessageTypes.TYPE_ERROR, e);
		}
		
		return cds;
	}

    public synchronized Map<Long, Role> getRoles() throws SRRoleNotFoundException {
	    List<Long> roleIDs = new LinkedList<Long>();

		try {
		    PreparedStatement st = conn.prepareStatement("SELECT ID FROM Roles ORDER BY Name");
		    try {
		    	if(st.execute()) {
		    		ResultSet rs = st.getResultSet();
		    		try {
		    			while (rs.next()) {
		    				roleIDs.add(new Long(rs.getString("ID")));
		    			}
		    		}
		    		finally {
		    			rs.close();
		    		}
		    	}
		    }
		    finally {
		    	st.close();
		    }
		} catch (SQLException e) {
		    handleSQLException(e);
			errorMessage(l10n._("Error while retrieving crypto data."), EMessageTypes.TYPE_ERROR, e);
			
			return null;
		}
		
		Map<Long, Role> roles = new HashMap<Long, Role>();
	    for(Iterator<Long> i = roleIDs.iterator(); i.hasNext(); ) {
	        long id = i.next();
	    	roles.put(id, getRole(id));
	    }

	    return roles;
	}

	private Role getRole(final long id) throws SRRoleNotFoundException {
		boolean rFound = false;
		String rKeyFPR = "";
		String rName = "";
		String rDescription = "";
		boolean rReadonly = true;
		boolean rHidden = true;
		String rSignSKey = "";

        try {
		    PreparedStatement st = conn.prepareStatement("SELECT RKeyFPR, Name, Description, Readonly, Hidden, SignSKey FROM Roles WHERE ID = ?");
		    try {
		        st.setLong(1, id);
		        if (st.execute()) {
		            ResultSet rs = st.getResultSet();
		            try {
		                while (rs.next()) {
                    		// TODO: Change it back to .getLong("...") if the bug is found.
                    		rFound = true;
                    		rKeyFPR = rs.getString("RKeyFPR");
                    		rName = rs.getString("Name");
                    		rDescription = rs.getString("Description");
                    		rReadonly = rs.getBoolean("Readonly");
                    		rHidden = rs.getBoolean("Hidden");
                    		rSignSKey = rs.getString("SignSKey");
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
            finally {
                st.close();
            }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Error while retrieving role ID (") + id + ").", EMessageTypes.TYPE_ERROR, e);
        }

		if(!rFound) 
			throw new SRRoleNotFoundException(id);

		return new Role(
			id,
			getRoleKey(id, rKeyFPR),
    		rKeyFPR,
    		rName,
    		rDescription,
    		rReadonly,
    		rHidden,
    		rSignSKey
    	);
	}
	
	private RoleKey getRoleKey(final long roleID, final String rKeyFPR) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT PubKey, EncKey FROM RoleKeys WHERE RoleID = ? AND KeyFPR = ?");
		    try {
    	        st.setLong(1, roleID);
    	        st.setString(2, rKeyFPR);
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	while (rs.next()) {
                    		return new RoleKey(rKeyFPR, rs.getString("PubKey"), rs.getString("EncKey"));
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Error while retrieving role key FPR (") + rKeyFPR + ").", EMessageTypes.TYPE_ERROR, e);
        }

	    return null;
	}

	public synchronized void saveCryptoData(final String prevTStamp, final CryptoEntry ce, final CryptoData cd) throws SRDBSQLException {
	    PreparedStatement st;

	    try {
	        st = conn.prepareStatement("EXECUTE saveCryptoData ?, ?, ?, ?, ?, ?, ?");
	        st.setString(1, prevTStamp);
	        st.setTimestamp(2, cd.getDate());
//	        st.setString(2, cd.getDate());
	        st.setLong(3, ce.getId());
	        st.setString(4, cd.getEncryptedData());
	        st.setString(5, cd.getUKeyFPR());
	        st.setString(6, cd.getSignKeyMap().get(ECryptoPGP_KeyType.ROLE_KEY));
	        st.setString(7, cd.getSignKeyMap().get(ECryptoPGP_KeyType.USER_KEY));
        
	        st.execute();

	        // TODO: Verify the return code of exec saveCryptoData.
	    } catch (SQLException e) {
	        handleSQLException(e);
	        throw new SRDBSQLException(e);
        }
	}

	public synchronized void createCryptoEntry(final CryptoEntry ce) throws SRDBSQLException {
	    PreparedStatement st;

	    try {
	        st = conn.prepareStatement("INSERT INTO CryptoEntries (CustomerID, Subject, Description, RoleID, RKeyFPR, Exportable, Readonly, Hidden, UKeyFPR, SignRKey, SignUKey) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
	        st.setLong(1, ce.getCustomer().getId());
	        st.setString(2, ce.getSubject());
	        st.setString(3, ce.getDescription());
	        st.setLong(4, ce.getRole().getID());
	        st.setString(5, ce.getRKey_FPR());
	        st.setBoolean(6, ce.isExportable());
	        st.setBoolean(7, ce.isReadonly());
	        st.setBoolean(8, ce.isHidden());
	        st.setString(9, ce.getUKeyFPR());
	        st.setString(10, ce.getSignKeyMap().get(ECryptoPGP_KeyType.ROLE_KEY));
	        st.setString(11, ce.getSignKeyMap().get(ECryptoPGP_KeyType.USER_KEY));
        
	        st.execute();

	        // TODO: Verify the return code of exec createCryptoEntry.
	    } catch (SQLException e) {
	        handleSQLException(e);
            throw new SRDBSQLException(e);
        }
	}
	
	public synchronized void updateCryptoEntry(final CryptoEntry ce) throws SRDBSQLException {
	    PreparedStatement st;
	    
	    try {
            st = conn.prepareStatement("UPDATE CryptoEntries SET Subject = ?, Description = ?, Readonly = ?, Hidden = ?, SignRKey = ?, SignUKey = ? WHERE ID = ?");
            st.setString(1, ce.getSubject());
            st.setString(2, ce.getDescription());
            st.setBoolean(3, ce.isReadonly());
            st.setBoolean(4, ce.isHidden());
            st.setString(5, ce.getSignKeyMap().get(ECryptoPGP_KeyType.ROLE_KEY));
            st.setString(6, ce.getSignKeyMap().get(ECryptoPGP_KeyType.USER_KEY));
            
            st.setLong(7, ce.getId());
        
            st.execute();	        
	    } catch (SQLException e) {
	        handleSQLException(e);
            throw new SRDBSQLException(e);
        }
	}
	
	public synchronized SigningKey getSigningKey() {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT Serial, PubKey, SignMKey FROM SigningKeys ORDER BY Serial DESC");
		    try {
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	if (rs.next()) {
                    		// TODO: Change it back to .getLong("...") if the bug is found.
                    		return new SigningKey(new Long(rs.getString("Serial")), rs.getString("PubKey"), rs.getString("SignMKey"));
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Failed to retrieve signing key."), EMessageTypes.TYPE_ERROR, e);
        }

	    return null;
	}
	
	public synchronized List<UserRole> getUserRolesByFPR(final String fpr) {
		List<UserRole> uroles = new LinkedList<UserRole>();
		
		try {
			PreparedStatement st = conn.prepareStatement("SELECT UserKeyFPR, RoleID, RKeyFPR, SignSKey, SignRKey FROM UserRoles WHERE UserKeyFPR = ?");
			try {
				st.setString(1, fpr);
				if(st.execute()) {
					ResultSet rs = st.getResultSet();
					try {
						while (rs.next()) {
							uroles.add(
									new UserRole(
											rs.getString("UserKeyFPR"),
											new Long(rs.getString("RoleID")),
											rs.getString("RKeyFPR"),
											rs.getString("SignSKey"),
											rs.getString("SignRKey")
									)
							);
						}
					}
					finally {
						rs.close();
					}
				}
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error at retrieving list of user roles."), EMessageTypes.TYPE_ERROR, e);
			
			return null;
		}
		
		return uroles;
	}	
	
	public synchronized List<UserRole> getUserRolesByRole(final Role role) {
		List<UserRole> uroles = new LinkedList<UserRole>();
		
		try {
			PreparedStatement st = conn.prepareStatement("SELECT UserKeyFPR, RoleID, RKeyFPR, SignSKey, SignRKey FROM UserRoles WHERE RoleID = ?");
			try {
				st.setLong(1, role.getID());
				if(st.execute()) {
					ResultSet rs = st.getResultSet();
					try {
						while (rs.next()) {
							uroles.add(
									new UserRole(
											rs.getString("UserKeyFPR"),
											new Long(rs.getString("RoleID")),
											rs.getString("RKeyFPR"),
											rs.getString("SignSKey"),
											rs.getString("SignRKey")
									)
							);
						}
					}
					finally {
						rs.close();
					}
				}
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error at retrieving list of user roles."), EMessageTypes.TYPE_ERROR, e);
			
			return null;
		}
		
		return uroles;
	}
	
	public synchronized List<PublicKey> getPublicKeys() {
		List<PublicKey> pubkeys = new LinkedList<PublicKey>();
		
		try {
			PreparedStatement st = conn.prepareStatement("SELECT FPR, PubKey FROM PublicKeys");
			try {
				if(st.execute()) {
					ResultSet rs = st.getResultSet();
					try {
						while (rs.next()) {
							pubkeys.add(
									new PublicKey(
											rs.getString("FPR"),
											rs.getString("PubKey"),
											null
									)
							);
						}
					}
					finally {
						rs.close();
					}
				}
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error at retrieving public keys."), EMessageTypes.TYPE_ERROR, e);
			
			return null;
		}
		
		return pubkeys;
	}

	public static void loadJDBCClasses() {
		final String jdbcClasses[] = new String[] {
			"sun.jdbc.odbc.JdbcOdbcDriver",
			"net.sourceforge.jtds.jdbc.Driver"
		};
		
		for(int i = 0; i < jdbcClasses.length; i++) {
			try {
				Class.forName(jdbcClasses[i]);
			} catch (ClassNotFoundException e2) {
				/* ignore class not found */
			}
		}
	}

	public void setLuceneStoreContent(LuceneStore file, byte[] data, int len) {
		remLuceneStore(file);

		try {
	    	PreparedStatement st = conn.prepareStatement("DELETE FROM LuceneStore WHERE RoleID = ? AND Name = ?");
			try {
				st.setLong(1, file.getRole().getID());
				st.setString(2, file.getName());

				st.execute();
			}
			finally {
				st.close();
			}

			st = conn.prepareStatement("INSERT INTO LuceneStore (RoleID, Name, Value, Size, LastModified) VALUES (?, ?, ?, ?, ?)");
			try {
				st.setLong(1, file.getRole().getID());
				st.setString(2, file.getName());
				st.setBytes(3, data);
				st.setLong(4, len);
				st.setLong(5, System.currentTimeMillis());

				st.execute();
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error storing search index file."), EMessageTypes.TYPE_ERROR, e);
		}		
	}

	public byte[] getLuceneStoreContent(LuceneStore file) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT Value FROM LuceneStore WHERE RoleID = ? AND Name = ?");
		    try {
		    	st.setLong(1, file.getRole().getID());
		    	st.setString(2, file.getName());
		    	
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	if (rs.next()) {
                    		return rs.getBytes("Value");
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Failed to retrieve search index file."), EMessageTypes.TYPE_ERROR, e);
        }

	    return null;
	}

	public long getLuceneStoreSize(LuceneStore file) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT Size FROM LuceneStore WHERE RoleID = ? AND Name = ?");
		    try {
		    	st.setLong(1, file.getRole().getID());
		    	st.setString(2, file.getName());
		    	
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	if (rs.next()) {
							return new Long(rs.getString("Size"));
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage("Failed to retrieve search index file size.", EMessageTypes.TYPE_ERROR, e);
        }

	    return 0;
	}

	public void remLuceneStore(LuceneStore file) {
	    try {
	    	PreparedStatement st = conn.prepareStatement("DELETE FROM LuceneStore WHERE RoleID = ? AND Name = ?");
			try {
				st.setLong(1, file.getRole().getID());
				st.setString(2, file.getName());

				st.execute();
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error removing search index file."), EMessageTypes.TYPE_ERROR, e);
		}		
	}

	public boolean existsLuceneStore(LuceneStore file) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT 1 FROM LuceneStore WHERE RoleID = ? AND Name = ?");
		    try {
		    	st.setLong(1, file.getRole().getID());
		    	st.setString(2, file.getName());
		    	
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	if (rs.next()) {
                    		return true;
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Failed test for search index file."), EMessageTypes.TYPE_ERROR, e);
        }

	    return false;
	}

	public long sizeLuceneStore(LuceneStore file) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT Size FROM LuceneStore WHERE RoleID = ? AND Name = ?");
		    try {
		    	st.setLong(1, file.getRole().getID());
		    	st.setString(2, file.getName());
		    	
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	if (rs.next()) {
                    		return rs.getLong("Size");
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Failed to retrieve search index file."), EMessageTypes.TYPE_ERROR, e);
        }

	    return 0;
	}

	public long modifiedLuceneStore(LuceneStore file) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT LastModified FROM LuceneStore WHERE RoleID = ? AND Name = ?");
		    try {
		    	st.setLong(1, file.getRole().getID());
		    	st.setString(2, file.getName());
		    	
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	if (rs.next()) {
                    		return rs.getLong("LastModified");
                    	}
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Failed to retrieve search index file."), EMessageTypes.TYPE_ERROR, e);
        }

	    return 0;
	}

	public String[] listLuceneStores(Role role) {
		try {
		    PreparedStatement st = conn.prepareStatement("SELECT Name FROM LuceneStore WHERE RoleID = ?");
		    try {
		    	st.setLong(1, role.getID());
		    	
    	        if (st.execute()) {
                    ResultSet rs = st.getResultSet();
                    try {
                    	List<String> ret = new ArrayList<String>();
                    	
                    	while (rs.next()) {
                    		ret.add(rs.getString("Name"));
                    	}

                    	String[] a = new String[ret.size()];
                    	ret.toArray(a);

                    	return a;
                    }
                    finally {
                    	rs.close();
                    }
    	        }
            }
		    finally {
		        st.close();
		    }
	    } catch (SQLException e) {
	        handleSQLException(e);
	        errorMessage(l10n._("Failed to retrieve search index file."), EMessageTypes.TYPE_ERROR, e);
        }

	    return null;
	}

	public void touchLuceneStore(LuceneStore file) {
		try {
	    	PreparedStatement st = conn.prepareStatement("INSERT INTO LuceneStore (RoleID, Name, Size, LastModified) VALUES (?, ?, ?, ?)");
			try {
				st.setLong(1, file.getRole().getID());
				st.setString(2, file.getName());
				st.setLong(3, 0);
				st.setLong(4, 1);

				st.execute();
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error storing search index file."), EMessageTypes.TYPE_ERROR, e);
		}		
	}

	public void clearLuceneStore(Role role) {
		try {
	    	PreparedStatement st = conn.prepareStatement("DELETE FROM LuceneStore WHERE RoleID = ?");
			try {
				st.setLong(1, role.getID());

				st.execute();
			}
			finally {
				st.close();
			}
		} catch (SQLException e) {
		    handleSQLException(e);
		    errorMessage(l10n._("Error clearing search index."), EMessageTypes.TYPE_ERROR, e);
		}		
	}
}

