//---------------------------------------------------------------------------
// Freely available from Media Art Online (http://www.media-art-online.org/).
// Copyright (C) 2003, 2004 Media Art Online (cafe@media-art-online.org)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// PROJECT: PGP/GnuPG Applications
// MODULE:  GnuPG support
// FILE:    GnuPG.java
//
// DESCRIPTION:
//
// REVISION HISTORY:
//  Name      Date        Description
//  ----      ----------  -----------
//  ks91      2002/11/05  initial version
//
// $Id: //depot/scm/main/java/org/media_art_online/gnupg/GnuPG.java#38 $
// $DateTime: 2007/09/23 18:11:18 $
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Package
//---------------------------------------------------------------------------
package org.media_art_online.gnupg;

//---------------------------------------------------------------------------
// Import
//---------------------------------------------------------------------------
import java.io.*;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.*;
import java.util.TreeSet;
import java.util.Vector;

/**
 * Class GnuPG is an abstraction of GnuPG software.
 *
 * @author <a href="mailto:ks91@media-art-online.org">Kenji Saito</a>
 */
public class GnuPG {

	public static final int PUBKEY_ALGO_RSA       =  1;
	public static final int PUBKEY_ALGO_RSA_E     =  2;
	public static final int PUBKEY_ALGO_RSA_S     =  3;
	public static final int PUBKEY_ALGO_ELGAMAL_E = 16;
	public static final int PUBKEY_ALGO_DSA       = 17;
	public static final int PUBKEY_ALGO_ELGAMAL   = 20;

	public static final int SECKEY = 1;
	public static final int SSBKEY = 2;
	public static final int PUBKEY = 3;
	public static final int SUBKEY = 4;

	public static final int RESULT_NOT_INSTALLED = -1;
	public static final int RESULT_OK            = 0;

	public static final int RESULT_SIG_CREATED    = 0;
	public static final int RESULT_SIG_TERMINATED = 1;
	public static final int RESULT_SIG_ALREADY    = 2;

	public static final int RESULT_DECRYPTION_OK  = 0;
	public static final int RESULT_DECRYPTION_NG  = 1;

	public static final int RESULT_NOSIG   = -1;
	public static final int RESULT_GOODSIG = 0;
	public static final int RESULT_BADSIG  = 1;
	public static final int RESULT_ERRSIG  = 2;

	public static final int RESULT_NO_SECKEY = 1;

	public static final int REASON_OK             = 0;
	public static final int REASON_EXPSIG         = 1;
	public static final int REASON_EXPKEYSIG      = 2;
	public static final int REASON_ALGORITHM      = 3;
	public static final int REASON_NO_PUBKEY      = 4;
	public static final int REASON_BAD_PASSPHRASE = 5;

	public static final int TRUST_BASE = 3;

	public static final int TRUST_REVOKED   = -3;
	public static final int TRUST_EXPIRED   = -2;
	public static final int TRUST_DISABLED  = -1;
	public static final int TRUST_INVALID   = 0;
	public static final int TRUST_UNDEFINED = 1;
	public static final int TRUST_NEVER     = 2;
	public static final int TRUST_MARGINAL  = 3;
	public static final int TRUST_FULL      = 4;
	public static final int TRUST_ULTIMATE  = 5;

	public static final String S_ARMOR_BEGIN = "-----BEGIN PGP SIGNATURE-----";
	public static final String S_ARMOR_END   = "-----END PGP SIGNATURE-----";
	public static final String S_ARMOR_SENTINEL	= "-----";

	public static final String S_EXT_ASC = ".asc";
	public static final String S_EXT_GPG = ".gpg";
	public static final String S_EXT_SIG = ".sig";

	public static final String S_GNUPG = "gnupg";

	public synchronized static void addSystemKey(GnuPGSystemKey key) {

		if (_loUnimported == null) {
			_loUnimported = new LinkedList<GnuPGSystemKey>();
		}

		for (Iterator<GnuPGSystemKey> iterator = _loUnimported.iterator();
		 iterator.hasNext(); ) {

			if ((iterator.next()).getFingerprint().equals(
			 key.getFingerprint())) {
				return;
			}
		}

		if (!key.isImported()) {
			_loUnimported.add(key);
		}
	}

	/**
	 * Adds an update listener.
	 *
	 * @param listener Update listener.
	 */
	public synchronized static void addUpdateListener(
	 GnuPGUpdateListener listener) {

		if (_manUpdate == null) {
			_manUpdate = new GnuPGUpdateManager();
		}

		_manUpdate.addListener(listener);
	}

	/**
	 * Checks signatures to a key.
	 *
	 * @param user GnuPG user (the whole key).
	 *
	 * @return Return value.
	 */
	public synchronized static int checkSignatures(GnuPGUser user) {
		_sDirPhoto     = null;
		_iteratorNames = (_user = user).iterator();
		_user.setSignatureChecked(true);

		return (command(null, null, "--check-sigs 0x"
		 + user.getPrimaryKey().getFingerprint(), ""));
	}

	/**
	 * Checks signatures to a key and its photo IDs.
	 *
	 * @param user      GnuPG user (the whole key).
	 * @param sDirPhoto Directory for photo-ID files.
	 *
	 * @return Return value.
	 */
	public synchronized static int checkSignatures(GnuPGUser user,
	 String sDirPhoto) {

		_sDirPhoto     = sDirPhoto;
		_iteratorNames = (_user = user).iterator();
		_user.setSignatureChecked(true);
		_indexPhoto    = 1;

		GnuPGPhoto.clearIcons(user, _sDirPhoto);

		return (command(null, null, "--show-photos --check-sigs 0x"
		 + user.getPrimaryKey().getFingerprint(), ""));
	}

	/**
	 * Decrypts a file.
	 *
	 * @param input       Input structure.
 	 * @param sPathOutput Path string of the output file.
	 * @param sPath       Path string of the file to be decrypted.
	 *
	 * @return Return value.
	 */
	public static int decrypt(GnuPGOutput output, GnuPGInput input,
	 String sPathOutput, String sPath) {
		return (command(output, input, "--decrypt -o " + sPathOutput, sPath));
	}

	/**
	 * Deletes a key.
	 *
	 * @param sUID       User ID of the key to delete.
	 *
	 * @return Return value.
	 */
	public static int deleteKey(String sUID) {
		return (command(null, null, "--delete-key " + sUID, ""));
	}

	/**
	 * Deletes a secret key.
	 *
	 * @param sUID       User ID of the key to delete.
	 *
	 * @return Return value.
	 */
	public static int deleteSecretKey(String sUID) {
		return (command(null, null, "--delete-secret-key " + sUID, ""));
	}

	/**
	 * Deletes a user ID.
	 *
	 * @param sUID    User ID of the key.
	 * @param index   Index for the deleting user ID.
	 *
	 * @return Return value.
	 */
	public static int deleteUserID(String sUID, int index) {

		return (editKey(new GnuPGEditInput(
		 new String[] {
			String.valueOf(index),
			"deluid",
			"save"
		 }), sUID));
	}

	/**
	 * Disables a key.
	 *
	 * @param sUID    User ID of the key.
	 *
	 * @return Return value.
	 */
	public static int disableKey(String sUID) {

		return (editKey(new GnuPGEditInput(
		 new String[] {
			"disable",
		 }), sUID));
	}

	/**
	 * Edits information of a key.
	 *
	 * @param input Input structure.
	 * @param sUID  Key to edit.
	 *
	 * @return Return value.
	 */
	public static int editKey(GnuPGInput input, String sUID) {
		return (command(null, input, "--edit-key " + sUID, ""));
	}

	/**
	 * Eables a key.
	 *
	 * @param sUID    User ID of the key.
	 *
	 * @return Return value.
	 */
	public static int enableKey(String sUID) {

		return (editKey(new GnuPGEditInput(
		 new String[] {
			"enable",
		 }), sUID));
	}

	/**
	 * Encrypts a file for a single recipient.
	 *
	 * @param sPath  Path string of the file to be encrypted.
	 * @param sUID   User ID of the recipient.
	 *
	 * @return Return value.
	 */
	public static int encrypt(String sPath, String sUID) {
		return (command(null, null, "--encrypt --recipient " + sUID, sPath));
	}

	public static int encrypt(String sPath, Collection<String> collection) {
		return (encrypt(null, sPath, collection, ""));
	}

	public static int encrypt(GnuPGInput input, String sPath,
	 Collection<String> collection, String sOption) {

		String sUIDs = "";

		for (Iterator<String> iterator = collection.iterator(); iterator.hasNext(); ) {
			sUIDs += "--recipient " + iterator.next() + " ";
		}

		return (command(null, input, "--encrypt " + sOption + " " + sUIDs,
		 sPath));
	}

	/**
	 * Encrypts a file for a single recipient in ASCII armor.
	 *
	 * @param sPath  Path string of the file to be encrypted.
	 * @param sUID   User ID of the recipient.
	 *
	 * @return Return value.
	 */
	public static int encryptArmor(String sPath, String sUID) {
		return (command(null, null, "--encrypt --armor --recipient "
		 + sUID, sPath));
	}

	public static int encryptSymmetric(String sPath, String sPhrase) {
		return (command(null, new GnuPGPhraseInput(sPhrase), "--symmetric ",
		 sPath));
	}

	/**
	 * Exports a key to a given file.
	 *
	 * @param sPath  Path string of the output file.
	 * @param sUID   User ID of the exporting key.
	 *
	 * @return Return value.
	 */
	public static int exportKey(String sPath, String sUID) {
		return (command(null, null,
		 "--export --output " + sPath + " " + sUID, ""));
	}

	/**
	 * Exports a secret key into a String.
	 *
	 * @param sUID   User ID of the exporting key.
	 *
	 * @return Returns the users key.
	 */
	public static String exportSecretKey(String sUID) {
		sUID = sUID.replaceAll(" ", "");

		return commandAsciiPipe(null, null,
				 "--export-secret-key --armor " + " " + sUID, "");
	}

	/**
	 * Exports a key to a given file in ASCII armor.
	 *
	 * @param sPath  Path string of the output file.
	 * @param sUID   User ID of the exporting key.
	 *
	 * @return Return value.
	 */
	public static int exportKeyArmor(String sPath, String sUID) {
		return (command(null, null,
		 "--export --armor --output " + sPath + " " + sUID, ""));
	}

	/**
	 * Fires an update event without actually updating GnuPG list of keys.
	 * This should be used for maintaining consistency within a GnuPG
	 * application (such as the choice of the secret key).
	 *
	 */
	public static void fireUpdate() {
		_manUpdate.fireUpdate();
	}

	/**
	 * Generates a key pair.
	 *
	 * @param input       Input structure.
	 *
	 * @return Return value.
	 */
	public static int generateKeyPair(GnuPGInput input) {
		return (command(null, input, "--gen-key", ""));
	}

	/**
	 * Generates a revocation certificate.
	 *
	 * @param input  Input structure.
	 * @param sPath  Path string of the output file.
	 * @param sUID   User ID of the key to generate the certificate for.
	 *
	 * @return Return value.
	 */
	public static int generateRevocationCertificate(GnuPGInput input,
	 String sPath, String sUID) {
		return (command(null, input,
		 "--gen-revoke --output " + sPath + " " + sUID, ""));
	}

	/**
	 * Gets a public key from its long ID.
	 *
	 * @param sID   Long ID of the primary key.
	 *
	 * @return Public key structure.
	 */
	public synchronized static GnuPGUser getUser(String sID) {
		return ((GnuPGUser)_ht.get(sID));
	}

	public synchronized static GnuPGVersion getVersion() {
		command(null, null, "--version", "");
		return (_version);
	}

	/**
	 * Imports a key from a given file.
	 *
	 * @param output GnuPG output structure.
	 * @param sPath  Path string of the input file.
	 *
	 * @return Return value.
	 */
	public static int importKey(GnuPGOutput output, String sPath) {

		output.clear();
		output.setOutputFor(GnuPGOutput.OUTPUT_IMPORT);
		return (command(output, null, "--import", sPath));
	}

	/**
	 * Returns the list of public keys maintained by this class.
	 *
	 * @return List of {@link GnuPGUser} objects representing public keys.
	 *
	 * @see GnuPGUser
	*/
	public synchronized static Set<GnuPGUser> listPublicKeys() {

		if (_setPublic == null) {
			_setPublic = new TreeSet<GnuPGUser>(new GnuPGComparator());
			_setPublic.addAll(_loPublic);
			_loPublic = null;
		}

		return (_setPublic);
	}

	/**
	 * Returns the list of secret keys maintained by this class.
	 *
	 * @return List of {@link GnuPGUser} objects representing secret keys.
	 *
	 * @see GnuPGUser
	*/
	public synchronized static Set<GnuPGUser> listSecretKeys() {

		if (_setSecret == null) {
			_setSecret = new TreeSet<GnuPGUser>(new GnuPGComparator());
			_setSecret.addAll(_loSecret);
			_loSecret = null;
		}

		return (_setSecret);
	}

	public synchronized static List<GnuPGSystemKey> listUnimportedSystemKeys() {

		if (_loUnimported == null) {
			_loUnimported = new LinkedList<GnuPGSystemKey>();
		}

		return (_loUnimported);
	}	

	/**
	 * Receives a key from the given server.
	 *
	 * @param output GnuPG output structure.
	 * @param sServer Domain name of the server.
	 * @param sUID    User ID of the key to receive.
	 *
	 * @return Return value.
	 */
	public static int receiveKey(GnuPGOutput output, String sServer,
	 String sUID) {

		output.clear();
		output.setOutputFor(GnuPGOutput.OUTPUT_IMPORT);
		return (command(output, null,
		 "--keyserver " + sServer + " --recv-keys " + sUID, ""));
	}

	/**
	 * Refreshes a key from a given file.
	 *
	 * @param output GnuPG output structure.
	 * @param sPath  Path string of the input file.
	 *
	 * @return Return value.
	 */
	public static int refreshKey(GnuPGOutput output, String sPath) {

		output.clear();
		output.setOutputFor(GnuPGOutput.OUTPUT_REFRESH);
		return (command(output, null, "--import", sPath));
	}

	/**
	 * Refreshes a key by the updates from the given server.
	 *
	 * @param output GnuPG output structure.
	 * @param sServer Domain name of the server.
	 * @param sUID    User ID of the key to receive.
	 *
	 * @return Return value.
	 */
	public static int refreshKey(GnuPGOutput output, String sServer,
	 String sUID) {

		output.clear();
		output.setOutputFor(GnuPGOutput.OUTPUT_REFRESH);
		return (command(output, null,
		 "--keyserver " + sServer + " --refresh-keys " + sUID, ""));
	}

	/**
	 * Searches a key from the given server.
	 *
	 * @param output  GnuPG output structure.
	 * @param input   Input structure.
	 * @param sServer Domain name of the server.
	 * @param sSearch Search string.
	 *
	 * @return Return value.
	 */
	public static int searchKey(GnuPGOutput output, GnuPGInput input,
	 String sServer, String sSearch) {

		output.clear();
		output.setOutputFor(GnuPGOutput.OUTPUT_IMPORT);
		return (command(output, input,
		 "--keyserver " + sServer + " --search-keys " + sSearch, ""));
	}

	/**
	 * Sends a key to the given server.
	 *
	 * @param sServer Domain name of the server.
	 * @param sUID    User ID of the sending key.
	 *
	 * @return Return value.
	 */
	public static int sendKey(String sServer, String sUID) {
		return (command(null, null,
		 "--keyserver " + sServer + " --send-keys " + sUID, ""));
	}

	public synchronized static void setLogger(Logger logger) {
		_logger = logger;
	}

	/**
	 * Sets the absolute path for gpg command.
	 *
	 * @param sPath Command path.
	 */
	public synchronized static void setPath(String sPath) {
		_sPathGpg = sPath;
	}

	/**
	 * Sets the progress monitor.
	 *
	 * @param monitor Progress monitor.
	 */
	public synchronized static void setProgressMonitor(
	 GnuPGProgressMonitor monitor) {
		_monitor = monitor;
	}

	/**
	 * Sets the owner trust of a key.
	 *
	 * @param sUID    User ID of the key.
	 * @param iTrust  Trust value.
	 *
	 * @return Return value.
	 */
	public static int setTrust(String sUID, int iTrust) {

		return (editKey(new GnuPGEditInput(
		 new String[] {
			"trust",
			String.valueOf(iTrust)
		 }), sUID));
	}

	/**
	 * Makes a detached signature.
	 *
	 * @param input  Input structure.
	 * @param sPath  Path string of the file to be signed.
	 * @param sUID   User ID of the signing key.
	 *
	 * @return One of the following:
	 * <table>
	 *  <tr><td>{@link #RESULT_SIG_CREATED}</td>
	 *   <td>Signature is created.</td></tr>
	 *  <tr><td>{@link #RESULT_SIG_TERMINATED}</td>
	 *   <td>Signature is not created.</td></tr>
	 * </table>
	 */
	public static int sign(GnuPGInput input, String sPath, String sUID) {
		return (command(null, input, "--detach-sign -u " + sUID, sPath));
	}

	public static int signAndEncrypt(GnuPGInput input, String sPath,
	 String sUID, Collection<String> collection) {
		return (encrypt(input, sPath, collection, "--sign -u " + sUID));
	}

	/**
	 * Makes a detached signature in ASCII armor.
	 *
	 * @param input  Input structure.
	 * @param sPath  Path string of the file to be signed.
	 * @param sUID   User ID of the signing key.
	 *
	 * @return One of the following:
	 * <table>
	 *  <tr><td>{@link #RESULT_SIG_CREATED}</td>
	 *   <td>Signature is created.</td></tr>
	 *  <tr><td>{@link #RESULT_SIG_TERMINATED}</td>
	 *   <td>Signature is not created.</td></tr>
	 * </table>
	 */
	public static int signArmor(GnuPGInput input, String sPath, String sUID) {
		return (command(null, input, "--detach-sign --armor -u " + sUID,
		 sPath));
	}

	public static int signKey(GnuPGInput input, GnuPGUser user,
	 String sUIDSelf) {
		return (signKey(input, "0x" + user.getPrimaryKey().getFingerprint(),
		 sUIDSelf));
	}

	/**
	 * Signs the given key.
	 *
	 * @param input      GnuPG Input structure.
	 * @param sUID       User ID of the key to be signed.
	 * @param sUIDSelf   User ID of the key to sign with.
	 *
	 * @return One of the following:
	 * <table>
	 *  <tr><td>{@link #RESULT_SIG_CREATED}</td>
	 *   <td>Signature is created.</td></tr>
	 *  <tr><td>{@link #RESULT_SIG_TERMINATED}</td>
	 *   <td>Signature is not created.</td></tr>
	 * </table>
	 */
	public static int signKey(GnuPGInput input, String sUID,
	 String sUIDSelf) {

		return (command(null, input, "--sign-key -u " + sUIDSelf
		 + " " + sUID, ""));
	}

	/**
	 * Updates the public/secret key lists maintained by this class.
	 *
	 * @return One of the following:
	 * <table>
	 *  <tr><td>{@link #RESULT_OK}</td>
	 *   <td>Key lists are updated.</td></tr>
	 *  <tr><td>{@link #RESULT_NOT_INSTALLED}</td>
	 *   <td>GnuPG is not installed on this machine.</td></tr>
	 * </table>
	 */
	public synchronized static int updateKeys() {

		int iResult = RESULT_OK;

		if (_ht == null) {
			_ht = new Hashtable<String, GnuPGUser>();
		}

		_setPublic = null;
		_setSecret = null;
		_user      = null;

		_loPublic = new LinkedList<GnuPGUser>();
		command(null, null,
		 "--with-colons --list-public-keys --fixed-list-mode"
		 + " --with-fingerprint --with-fingerprint", "");

		_loSecret = new LinkedList<GnuPGUser>();
		iResult = command(null, null,
		 "--with-colons --list-secret-keys --fixed-list-mode"
		 + " --with-fingerprint --with-fingerprint", "");

		for (Iterator<GnuPGUser> iterator = _loSecret.iterator(); iterator.hasNext(); ) {

			GnuPGUser userSecret = iterator.next();
			GnuPGUser userPublic
			 = (GnuPGUser)_ht.get(userSecret.getPrimaryKey().getID());

			if (userPublic != null) {
				userPublic.setPair(userSecret);
			}
		}

		if (iResult == RESULT_OK && _manUpdate != null) {
			_manUpdate.fireUpdate();
		}

		return (iResult);
	}

	/**
	 * Verifies a detached signature.
	 *
	 * @param output Output structure.
	 * @param sPath	 Path string of the signature file.
	 *
	 * @return One of the following:
	 * <table>
	 *  <tr><td>{@link #RESULT_GOODSIG}</td>
	 *   <td>Signature is good.</td></tr>
	 *  <tr><td>{@link #RESULT_BADSIG}</td>
	 *   <td>Signature is bad.</td></tr>
	 *  <tr><td>{@link #RESULT_ERRSIG}</td>
	 *   <td>Signature is erroneous.</td></tr>
	 * </table>
	 */
	public static int verify(GnuPGOutput output, String sPath) {

		output.clear();
		output.setOutputFor(GnuPGOutput.OUTPUT_VERIFY);
		return (command(output, null, "--verify", sPath));
	}

//===========================================================================
// BARRIER: Anything below is not open to other files.
//===========================================================================

	/**
	 * Internal command processor.
	 *
	 * @param output   Output structure.
	 * @param input    Input structure.
	 * @param sCommand command string.
	 * @param sPath    command parameter (usually the path string for a file).
	 *
	 * @return result code.
	 */
	private synchronized static String commandAsciiPipe(GnuPGOutput output,
			 GnuPGInput input, String sCommand, String sPath) {

		try {
			Process process = Runtime.getRuntime().exec(
			 _sPathGpg + " --status-fd 1 --command-fd 0 --no-tty " + sCommand
			 + " " + sPath);

			BufferedReader out = new BufferedReader(new InputStreamReader(
					 process.getInputStream(), "UTF-8"));

			String res = "";
			String l;
			while((l = out.readLine()) != null) {
				res += l + System.getProperty("line.separator");
			}

			return res;
		} catch (IOException exIO) {

		} catch (Throwable exThrowable) {
			exThrowable.printStackTrace();
		}

		return null;
	}


	/**
	 * Internal command processor.
	 *
	 * @param output   Output structure.
	 * @param input    Input structure.
	 * @param sCommand command string.
	 * @param sPath    command parameter (usually the path string for a file).
	 *
	 * @return result code.
	 */
	private synchronized static int command(GnuPGOutput output,
	 GnuPGInput input, String sCommand, String sPath) {

		List<?> list;
		String s;
		int iResult = RESULT_OK;

		BufferedReader reader = null;
		PrintWriter    writer = null;

		try {
			Process process = Runtime.getRuntime().exec(
			 _sPathGpg + " --status-fd 1 --command-fd 0 --no-tty " + sCommand
			 + " " + sPath);

			writer = new PrintWriter(process.getOutputStream(), true);
			reader = new BufferedReader(new InputStreamReader(
			 process.getInputStream(), "UTF-8"));

			while ((s = reader.readLine()) != null) {

				if (_logger != null) {
					_logger.finest("gpg: " + s);
				}

				if (s.startsWith(S_VERSION)) {

					if (_version == null) {
						_version = new GnuPGVersion();
						_version.setVersionString(
						 s.substring(S_VERSION.length()));
					}
					continue;

				} else if (s.startsWith(S_HOME)) {
					_version.setHomeDirectory(s.substring(S_HOME.length()));
					continue;
				}

				list = parseResponse(s);

				switch (((Integer)list.get(0)).intValue()) {
					case TYPE_GNUPG:
						switch (((Integer)list.get(1)).intValue()) {
							case FC_GOODSIG:
								iResult = RESULT_GOODSIG;
								output.setPublicKeyName(getTailString(list,
								 CNT_ITEM_X_SIG),
								 (String)list.get(CNT_ITEM_LONG_ID));
								break;

							case FC_EXPSIG:
								output.setVerifiedReason(REASON_EXPSIG);
								break;

							case FC_EXPKEYSIG:
								output.setVerifiedReason(REASON_EXPKEYSIG);
								break;

							case FC_BADSIG:
								iResult = RESULT_BADSIG;
								output.setPublicKeyName(getTailString(list,
								 CNT_ITEM_X_SIG),
								 (String)list.get(CNT_ITEM_LONG_ID));
								break;

							case FC_ERRSIG:
								iResult = RESULT_ERRSIG;

								switch (((String)list.get(
								 CNT_ITEM_ERRSIG_RC)).charAt(0)) {
									case CASE_ERRSIG_ALGORITHM:
										output.setVerifiedReason(
										 REASON_ALGORITHM);
										break;

									case CASE_ERRSIG_NO_PUBKEY:
										output.setVerifiedReason(
										 REASON_NO_PUBKEY);
										break;

									default:
										break;
								}
								break;

							case FC_VALIDSIG:
								{
									output.setSignatureTime(
									 Long.parseLong((String)list.get(
									 CNT_ITEM_VALIDSIG_TIMESTAMP))
									 * 1000);
								}
								break;

							case FC_NODATA:
								iResult = RESULT_BADSIG;
								break;

							case FC_TRUST_UNDEFINED:
								output.setPublicKeyValidity(TRUST_UNDEFINED);
								break;

							case FC_TRUST_NEVER:
								output.setPublicKeyValidity(TRUST_NEVER);
								break;

							case FC_TRUST_MARGINAL:
								output.setPublicKeyValidity(TRUST_MARGINAL);
								break;

							case FC_TRUST_FULL:
								output.setPublicKeyValidity(TRUST_FULL);
								break;

							case FC_TRUST_ULTIMATE:
								output.setPublicKeyValidity(TRUST_ULTIMATE);
								break;

							case FC_GET_BOOL:
								{
									String sX
									 = (String)list.get(CNT_ITEM_GET_X);

									if (sX.equals("openfile.overwrite.okay")
									 || sX.equals("sign_uid.okay")
									 || sX.equals("keyedit.sign_all.okay")
									 || sX.equals("keyedit.remove.uid.okay")
									 || sX.equals("photoid.jpeg.size")
									 || sX.equals("delete_key.okay")
									 || sX.equals("delete_key.secret.okay")
									 || sX.equals("change_passwd.empty.okay")
									 || sX.equals("ask_revocation_reason.okay")
									 || sX.equals("gen_revoke.okay")) {

										if (writer != null) {
											writer.println("YES");
										}

									} else {

										if (writer != null) {
											writer.println("NO");
										}
									}
								}
								break;

							case FC_GET_HIDDEN:
								if (((String)list.get(CNT_ITEM_GET_X))
								 .equals("passphrase.enter")) {

									if (writer != null) {

										String sp = input.getPassphrase();

										if (sp == null) {
												return (RESULT_SIG_TERMINATED);

										} else {
											writer.println(sp);
										}
									}
								}
								break;

							case FC_GET_LINE:
								{
									if (writer != null) {

										String sl = input.getLine(
										 (String)list.get(CNT_ITEM_GET_X));

										if (sl == null) {
												return (RESULT_SIG_TERMINATED);

										} else {
											writer.println(sl);
										}
									}
								}
								break;

							case FC_BAD_PASSPHRASE:
								iResult = RESULT_SIG_TERMINATED;
								input.setReason(REASON_BAD_PASSPHRASE);
								break;

							case FC_USERID_HINT:
								input.setUserName(getTailString(list,
								 CNT_ITEM_USERID_HINT));
								break;

							case FC_ALREADY_SIGNED:
								input.setReason(RESULT_SIG_ALREADY);
								break;

							case FC_SIG_CREATED:
								iResult = RESULT_SIG_CREATED;
								break;

							case FC_IMPORTED:
								output.setPublicKeyName(getTailString(list,
								 CNT_ITEM_X_SIG),
								 (String)list.get(CNT_ITEM_LONG_ID));
								break;

							case FC_IMPORT_OK:
								output.setFingerprint(
								 (String)list.get(CNT_ITEM_FINGERPRINT));
								break;

							case FC_NO_SECKEY:
								iResult = RESULT_NO_SECKEY;
								break;

							case FC_PROGRESS:
								if (_monitor != null) {
									_monitor.makeProgress();
								}
								break;

							case FC_DECRYPTION_OK:
								iResult = RESULT_DECRYPTION_OK;
								break;

							case FC_DECRYPTION_NG:
								iResult = RESULT_DECRYPTION_NG;
								break;

							default:
								break;
						}
						break;

					case TYPE_SECKEY:
						if (_loSecret == null) {
							_loSecret = new LinkedList<GnuPGUser>();
						}

						_loSecret.add((GnuPGUser)list.get(1));
						break;

					case TYPE_PUBKEY:
						if (_loPublic == null) {
							_loPublic = new LinkedList<GnuPGUser>();
						}

						_loPublic.add((GnuPGUser)list.get(1));
						break;

					case TYPE_SEARCHED:
						input.setUserName((String)list.get(1));
						break;

					default:
						break;
				}
			}

		} catch (IOException exIO) {
System.out.println(exIO.toString());
			iResult = RESULT_NOT_INSTALLED;

		} catch (Throwable exThrowable) {
			exThrowable.printStackTrace();

		} finally {

			try {

				if (writer != null) {
					writer.close();
				}
				if (reader != null) {
					reader.close();
				}

			} catch (Exception unused) {
			}
		}

		return (iResult);
	}

	private static List<Object> parseResponse(String sResponse) {

		List<Object> list = new LinkedList<Object>();

		//
		// "[GNUPG:] KEYEXPIRED" appears whenever gpg detects key expiration.
		//
		if (sResponse.indexOf("KEYEXPIRED") >= 0) {
			list.add(new Integer(CODE_UNDEF));
			return (list);
		}

		int code;
		int idx = 0;
		int i;

		if ((i = sResponse.indexOf(S_DIV, idx)) >= 0 && i <= MAX_TYPE_LENGTH) {

			list.add(new Integer(code
			 = searchForCode(sResponse.substring(idx, i), _aoType)));

			if (code == TYPE_GNUPG) {
				if ((i = sResponse.indexOf(" ", idx = i + 3)) >= 0) {
					list.add(new Integer(code
					 = searchForCode(sResponse.substring(idx, i), _aoFc)));

					while ((i = sResponse.indexOf(" ", idx = i + 1)) >= 0) {
						list.add(sResponse.substring(idx, i));
					}

					list.add(sResponse.substring(idx));

				} else {
					list.add(new Integer(
					 searchForCode(sResponse.substring(idx), _aoFc)));
				}

			} else if (code != CODE_UNDEF) {

				for (int cnt = 1;
				 (i = sResponse.indexOf(S_DIV, idx = i + S_DIV.length())) >= 0;
				 cnt++) {
					list.add(sResponse.substring(idx, i));
				}

				switch (code) {
					case TYPE_SECKEY:
					case TYPE_PUBKEY:
					{
						GnuPGKey key = new GnuPGKey(
						 (Integer)list.get(CNT_FIELD_TYPE),
						 (String)list.get(CNT_FIELD_LENGTH_KEY),
						 (String)list.get(CNT_FIELD_ALGO_KEY),
						 (String)list.get(CNT_FIELD_ID_KEY),
						 (String)list.get(CNT_FIELD_DATE_CREATION),
						 (String)list.get(CNT_FIELD_DATE_EXPIRATION));

						_user = new GnuPGUser(code == TYPE_SECKEY
						 ? null : (String)list.get(CNT_FIELD_VALIDITY),
						 code == TYPE_SECKEY
						 ? null : (String)list.get(CNT_FIELD_TRUST));

						if (code != TYPE_SECKEY
						 && ((String)list.get(CNT_FIELD_STATUS))
						 .endsWith("D")) {
							_user.setValidity(TRUST_DISABLED);
						}

						_user.setPrimaryKey(key);
						list.add(1, _user);

						if (code == TYPE_PUBKEY) {
							_ht.put(key.getID(), _user);
						}
					}
						break;

					case TYPE_SUBKEY:
					case TYPE_SSBKEY:
					{
						if (_user == null) {
							/* subkey of an expired key or signature */
							/* is reported.                          */

						} else {

							GnuPGKey key = new GnuPGKey(
							 (Integer)list.get(CNT_FIELD_TYPE),
							 (String)list.get(CNT_FIELD_LENGTH_KEY),
							 (String)list.get(CNT_FIELD_ALGO_KEY),
							 (String)list.get(CNT_FIELD_ID_KEY),
							 (String)list.get(CNT_FIELD_DATE_CREATION),
							 (String)list.get(CNT_FIELD_DATE_EXPIRATION));

							_user.addSubkey(key);
						}
					}
						break;

					case TYPE_USER_ID:
					{
						if (_user == null) {
							/* user id of an expired key or signature */
							/* is reported.                           */

						} else {
							_user.addName(new GnuPGName(
							 (String)list.get(CNT_FIELD_ID_USER),
							 _user.getPrimaryKey().getID()));
						}
					}
						break;

					case TYPE_PHOTO_ID:
					{
						if (_user == null) {
							/* photo id of an expired key or signature */
							/* is reported.                           */

						} else {
							_user.addName(new GnuPGPhoto(
							 _user.getPrimaryKey().getID()));
						}
					}
						break;

					case TYPE_FINGERPRINT:
					{
						if (_user == null) {
							/* fingerprint of an expired key is reported. */

						} else {

							if (_user.getSubkeys().size() > 0) {

								((GnuPGKey)_user.getSubkeys().getLast())
								 .setFingerprint(
								 (String)list.get(CNT_FIELD_FINGERPRINT));

							} else {
								_user.getPrimaryKey().setFingerprint(
								 (String)list.get(CNT_FIELD_FINGERPRINT));
							}
						}
					}
						break;

					default:
						break;
				}
			}

		} else if (sResponse.startsWith("sig")) {

			if (_vectorSigs != null) {
				_vectorSigs.add(sResponse.charAt(COL_SIG) == ' '
				 ? sResponse.substring(COL_SIG + 1)
				 : sResponse.substring(COL_SIG));
			}
			list.add(new Integer(CODE_UNDEF));

		} else if (sResponse.startsWith("pub")
		 || sResponse.startsWith("uid")) {

			list.add(new Integer(CODE_UNDEF));

			if (_vectorSigs != null) {

				if (_vectorSigs.size() <= 0) {
					return (list);
				}

				_name.setSignatures(_vectorSigs);
				_vectorSigs = null;
			}

			String s = sResponse.substring(COL_SIG_USER_ID);

			if (!s.startsWith("[revoked]") && _iteratorNames != null
			 && _iteratorNames.hasNext()) {
				_name = (GnuPGName)_iteratorNames.next();

				if (_sDirPhoto != null
				 && _name.getClass() == GnuPGPhoto.class) {

					((GnuPGPhoto)_name).createIcon(_user, _sDirPhoto,
					 _indexPhoto++);
				}

				_vectorSigs = new Vector<String>();
			}

		} else if (sResponse.startsWith("sub")) {

			if (_name != null && _vectorSigs != null) {
				_name.setSignatures(_vectorSigs);
				_vectorSigs = null;
			}

			list.add(new Integer(CODE_UNDEF));

		} else if (sResponse.length() <= 0) {

			list.add(new Integer(CODE_UNDEF));

		} else if (sResponse.startsWith(S_MARK_SEARCH_RESULT_HEADING)
		 || sResponse.charAt(0) == '('
		 && Character.isDigit(sResponse.charAt(1))
		 || Character.isWhitespace(sResponse.charAt(0))
		 && Character.isWhitespace(sResponse.charAt(1))) {

			list.add(new Integer(TYPE_SEARCHED));
			list.add(sResponse);

		} else {
			list.add(new Integer(CODE_UNDEF));
		}

		return (list);
	}

	private static String getTailString(List<?> list, int cnt) {
		String s = null;

		while (cnt < list.size()) {

			s = (s == null
			 ? (String)list.get(cnt) : (s + " " + (String)list.get(cnt)));

			cnt++;
		}

		return (s);
	}

	private static int searchForCode(String s, GnuPGCode[] aog) {
		int i;
		int rc;
		int min = 0;
		int max = aog.length - 1;

		while (min <= max) {
			i = (min + max) / 2;

			if ((rc = s.compareTo(aog[i].s)) == 0) {
				return (aog[i].iCode);

			} else if (rc < 0) {
				max = i - 1;

			} else {
				min = i + 1;
			}
		}

		return (CODE_UNDEF);
	}

	private static Hashtable<String, GnuPGUser> _ht = null;

	private static Iterator<GnuPGName> _iteratorNames = null;

	private static LinkedList<GnuPGUser> _loPublic = null;
	private static LinkedList<GnuPGUser> _loSecret = null;

	private static LinkedList<GnuPGSystemKey> _loUnimported = null;

	private static Logger _logger = null;

	private static GnuPGName _name = null;

	private static GnuPGProgressMonitor _monitor = null;

	private static GnuPGUpdateManager _manUpdate = null;

	private static GnuPGUser _user = null;

	private static GnuPGVersion _version = null;

	private static String _sDirPhoto = null;

	private static String _sPathGpg = "";

	private static Set<GnuPGUser> _setPublic = null;
	private static Set<GnuPGUser> _setSecret = null;

	private static Vector<String> _vectorSigs = null;

	private static int _indexPhoto = 0;

	private static final String S_HOME = "Home: ";

	private static final String S_MARK_SEARCH_RESULT_HEADING = "Keys";

	private static final String S_VERSION = "gpg (GnuPG) ";

	private static final int CODE_UNDEF  = -1;

	private static final int TYPE_GNUPG       = 0;

	private static final int TYPE_SECKEY      = 1;
	private static final int TYPE_SSBKEY      = 2;
	private static final int TYPE_PUBKEY      = 3;
	private static final int TYPE_SUBKEY      = 4;

	private static final int TYPE_USER_ID     = 5;
	private static final int TYPE_PHOTO_ID    = 6;
	private static final int TYPE_FINGERPRINT = 7;

	private static final int TYPE_SEARCHED    = 8;

	private static final int FC_GOODSIG         =  1;
	private static final int FC_EXPSIG          =  2;
	private static final int FC_EXPKEYSIG       =  3;
	private static final int FC_BADSIG          =  4;
	private static final int FC_ERRSIG          =  5;
	private static final int FC_VALIDSIG        =  6;
	/*private static final int FC_SIG_ID          =  7;*/
	private static final int FC_ENC_TO          =  8;
	private static final int FC_NODATA          =  9;
	private static final int FC_UNEXPECTED      = 10;
	private static final int FC_TRUST_UNDEFINED = 11;
	private static final int FC_TRUST_NEVER     = 12;
	private static final int FC_TRUST_MARGINAL  = 13;
	private static final int FC_TRUST_FULL      = 14;
	private static final int FC_TRUST_ULTIMATE  = 15;

	private static final int FC_GET_BOOL        = 16;
	private static final int FC_GET_HIDDEN      = 17;
	private static final int FC_GET_LINE        = 18;

	private static final int FC_BAD_PASSPHRASE  = 19;

	private static final int FC_USERID_HINT     = 20;

	private static final int FC_ALREADY_SIGNED  = 21;
	private static final int FC_SIG_CREATED     = 22;

	private static final int FC_IMPORTED        = 23;
	private static final int FC_IMPORT_OK       = 24;

	private static final int FC_NO_SECKEY       = 25;

	private static final int FC_PROGRESS        = 26;

	private static final int FC_DECRYPTION_OK   = 27;
	private static final int FC_DECRYPTION_NG   = 28;

	private static final int CNT_ITEM_GET_X              = 2;
	private static final int CNT_ITEM_FINGERPRINT        = 3;
	private static final int CNT_ITEM_USERID_HINT        = 3;
	private static final int CNT_ITEM_LONG_ID            = 2;
	private static final int CNT_ITEM_X_SIG              = 3;
	private static final int CNT_ITEM_VALIDSIG_TIMESTAMP = 4;
	private static final int CNT_ITEM_ERRSIG_RC          = 7;

	private static final int CNT_FIELD_TYPE            = 0;
	private static final int CNT_FIELD_VALIDITY        = 1;
	private static final int CNT_FIELD_LENGTH_KEY      = 2;
	private static final int CNT_FIELD_ALGO_KEY        = 3;
	private static final int CNT_FIELD_ID_KEY          = 4;
	private static final int CNT_FIELD_DATE_CREATION   = 5;
	private static final int CNT_FIELD_DATE_EXPIRATION = 6;
	private static final int CNT_FIELD_TRUST           = 8;
	private static final int CNT_FIELD_ID_USER         = 9;
	private static final int CNT_FIELD_FINGERPRINT     = 9;
	private static final int CNT_FIELD_STATUS          = 11;

	private static final int COL_SIG         = 12;
	private static final int COL_SIG_USER_ID = 31;

	private static final int MAX_TYPE_LENGTH = 6;

	private static final char CASE_ERRSIG_ALGORITHM = '4';
	private static final char CASE_ERRSIG_NO_PUBKEY = '9';

	private static final String S_DIV = ":";

	private static final GnuPGCode[] _aoType = {
		new GnuPGCode("[GNUPG", TYPE_GNUPG),
		new GnuPGCode("fpr",    TYPE_FINGERPRINT),
		new GnuPGCode("pub",    TYPE_PUBKEY),
		new GnuPGCode("sec",    TYPE_SECKEY),
		new GnuPGCode("ssb",    TYPE_SSBKEY),
		new GnuPGCode("sub",    TYPE_SUBKEY),
		new GnuPGCode("uat",    TYPE_PHOTO_ID),
		new GnuPGCode("uid",    TYPE_USER_ID)
	};

	private static final GnuPGCode[] _aoFc = {
		new GnuPGCode("ALREADY_SIGNED",    FC_ALREADY_SIGNED),
		new GnuPGCode("BADSIG",            FC_BADSIG),
		new GnuPGCode("BAD_PASSPHRASE",    FC_BAD_PASSPHRASE),
		new GnuPGCode("DECRYPTION_FAILED", FC_DECRYPTION_NG),
		new GnuPGCode("DECRYPTION_OKAY",   FC_DECRYPTION_OK),
		new GnuPGCode("ENC_TO",            FC_ENC_TO),
		new GnuPGCode("ERRSIG",            FC_ERRSIG),
		new GnuPGCode("EXPKEYSIG",         FC_EXPKEYSIG),
		new GnuPGCode("EXPSIG",            FC_EXPSIG),
		new GnuPGCode("GET_BOOL",          FC_GET_BOOL),
		new GnuPGCode("GET_HIDDEN",        FC_GET_HIDDEN),
		new GnuPGCode("GET_LINE",          FC_GET_LINE),
		new GnuPGCode("GOODSIG",           FC_GOODSIG),
		new GnuPGCode("IMPORTED",          FC_IMPORTED),
		new GnuPGCode("IMPORT_OK",         FC_IMPORT_OK),
		new GnuPGCode("NODATA",            FC_NODATA),
		new GnuPGCode("NO_SECKEY",         FC_NO_SECKEY),
		new GnuPGCode("PROGRESS",          FC_PROGRESS),
		new GnuPGCode("SIG_CREATED",       FC_SIG_CREATED),
		new GnuPGCode("TRUST_FULLY",       FC_TRUST_FULL),
		new GnuPGCode("TRUST_MARGINAL",    FC_TRUST_MARGINAL),
		new GnuPGCode("TRUST_NEVER",       FC_TRUST_NEVER),
		new GnuPGCode("TRUST_ULTIMATE",    FC_TRUST_ULTIMATE),
		new GnuPGCode("TRUST_UNDEFINED",   FC_TRUST_UNDEFINED),
		new GnuPGCode("UNEXPECTED",        FC_UNEXPECTED),
		new GnuPGCode("USERID_HINT",       FC_USERID_HINT),
		new GnuPGCode("VALIDSIG",          FC_VALIDSIG)
	};
}

/**
 * Internal class representing a (string, code) pair.
 *
 * @author <a href="mailto:ks91@media-art-online.org">Kenji Saito</a>
 */
class GnuPGCode {

	/**
	 * Constructor.
	 *
	 * @param s     GnuPG response string.
	 * @param iCode org.media_art_online.gnupg internal code.
	 *
	 * @see org.media_art_online.gnupg.GnuPG#_aoType
	 * @see org.media_art_online.gnupg.GnuPG#_aoFc
	*/
	public GnuPGCode(String s, int iCode) {
		this.s = s;
		this.iCode = iCode;
	}

	/**
	 * GnuPG response string.
	 */
	public final String s;

	/**
	 * org.media_art_online.gnupg internal code.
	 */
	public final int iCode;
}

class GnuPGUpdateManager {

	public GnuPGUpdateManager() {

		_loListeners = new LinkedList<GnuPGUpdateListener>();
	}

	public synchronized void addListener(GnuPGUpdateListener listener) {
		_loListeners.add(listener);
	}

	public synchronized void fireUpdate() {

		if (_loListeners != null) {

			for (Iterator<GnuPGUpdateListener> iterator = _loListeners.iterator();
			 iterator.hasNext(); ) {
				iterator.next().gnupgUpdated();
			}
		}
	}

	private List<GnuPGUpdateListener> _loListeners;
}

// end of GnuPG.java
